# usbhotkey script to handle the Alphagrip AG5
# Copyright (C) 2007  Lars Krueger
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

# We assume a UK keymap here. For a US keymap exchange " and @

print "usbhotkey AG5 script starting\n"

# This script places all mappings in ruby tables which are indexed by some
# state variable and the USB scan code. The backside keys are mapped to a
# normalised id first.

# The basic idea of this script is to provide three color levels: red, green,
# and blue, which can be pressed together to acchieve mixed colours. Thus, 8
# additional shift levels can be implemented. These levels can also generate
# different keypresses with respect to the state of the shift key.
#
# All keys can send X11 keycode sequences of arbitrary length.

# Printscreen on the AG5 quits the program

##################################### Keymap ###################################
#==================== simple keys map from USB to X11 direct ===================
# Index 0: unshifted, index 1: shifted
SimpleKeys = []
SimpleKeys[0]={
  13 => [ Win_Control_L ],
  25 => [ Win_Alt_L ],
  27 => [ Win_Alt_R ],
  29 => [ Win_Control_R ],
  15 => [ Win_Return ],
  44 => [ Win_space ]}

SimpleKeys[1]={
  13 => [ Win_Control_L ],
  25 => [ Win_Alt_L ],
  27 => [ Win_Alt_R ],
  29 => [ Win_Control_R ],
  15 => [ Win_Tab ],
  42 => [ Win_BackSpace ],
  44 => [ Win_Escape ]}

#======================== Set or release the colorshift ========================
ColorKeys = {
  6  => [0, 2],
  28 => [0, 4],
  14 => [0, 1],
  40 => [1, 4],
  42 => [1, 2],
  43 => [1, 1]}

#=========================== USB code to backside id ===========================
BacksideId={
  USB_g         => 0 ,
  USB_r         => 1 ,
  USB_f         => 2 ,
  USB_t         => 3 ,
  USB_d         => 4 ,
  USB_e         => 5 ,
  USB_w         => 6 ,
  USB_s         => 7 ,
  USB_q         => 8 ,
  USB_a         => 9 ,
  USB_m         => 10,
  USB_h         => 11,
  USB_n         => 12,
  USB_u         => 13,
  USB_i         => 14,
  USB_Comma     => 15,
  USB_o         => 16,
  USB_Period    => 17,
  USB_p         => 18,
  USB_b         => 19,
  }

#================================ Normal keymap ================================
# Dimension 1 by ORed colorshift values
# Dimension 2 shift state
# Dimension 3 by backside ID
KeyMap=[[], []]

# g|r|f|t|d|e|w|s|q|a|m|h|n|u|i|l|o|c|p|b"
KeyMap[0][0] = [ 
  [ Win_G],
  [ Win_R],
  [ Win_F],
  [ Win_T],
  [ Win_D],
  [ Win_E],
  [ Win_W],
  [ Win_S],
  [ Win_Q],
  [ Win_A],
  [ Win_M],
  [ Win_H],
  [ Win_N],
  [ Win_U],
  [ Win_I],
  [ Win_L],
  [ Win_O],
  [ Win_C],
  [ Win_P],
  [ Win_B]]

# {up}|{down}|{left}|{right}|{tab}|{space}|:|{del}|@|{esc}|{pgup}|{pgdn}|{home}|{end}|{enter}|{bs}|.|;|,|{printscreen}"
KeyMap[0][1] = [ 
  [ Win_Up],
  [ Win_Down],
  [ Win_Left],
  [ Win_Right],
  [ Win_Tab],
  [ Win_space],
  [ Win_Shift_L, Win_semicolon],
  [ Win_Delete],
  [ Win_Shift_L, Win_apostrophe],
  [ Win_Escape],
  [ Win_Page_Up],
  [ Win_Page_Down],
  [ Win_Home],
  [ Win_End],
  [ Win_Return],
  [ Win_BackSpace],
  [ ],
  [ Win_semicolon],
  [ ],
  [ Win_Print]]

# {#}|2|,|1|{+}|3|-|4|=|5|$|7|6|.|8|*|9|;|0|/"
KeyMap[0][2] = [ 
  [ Win_numbersign ],
  [ Win_2 ],
  [ Win_period ],
  [ Win_1 ],
  [ Win_Shift_L, Win_equal ],
  [ Win_3 ],
  [ Win_minus ],
  [ Win_4 ],
  [ Win_equal ],
  [ Win_5 ],
  [ Win_Shift_L, Win_4 ],
  [ Win_7 ],
  [ Win_6 ],
  [ Win_comma ],
  [ Win_8 ],
  [ Win_Shift_L, Win_8 ],
  [ Win_9 ],
  [ Win_semicolon ],
  [ Win_0 ],
  [ Win_slash ]]

# | | |j||v| |y| || | |k||x| |z|| | "
KeyMap[0][3] = [ 
  [  ],
  [  ],
  [  ],
  [ Win_J ],
  [  ],
  [ Win_V ],
  [  ],
  [ Win_Y ],
  [  ],
  [  ],
  [  ],
  [  ],
  [ Win_K ],
  [  ],
  [ Win_X ],
  [  ],
  [ Win_Z ],
  [  ],
  [  ],
  [  ]]

# ~,%,"",',(,),{{},{}},[,],{^},_,?,!,&,|,<,>,``,\"
KeyMap[0][4] = [ 
  [ Win_Shift_L, Win_numbersign ],
  [ Win_Shift_L, Win_5 ],
  [ Win_Shift_L, Win_2 ],
  [ Win_apostrophe],
  [ Win_Shift_L, Win_9 ],
  [ Win_Shift_L, Win_0 ],
  [ Win_Shift_L, Win_bracketleft ],
  [ Win_Shift_L, Win_bracketright ],
  [ Win_bracketleft  ],
  [ Win_bracketright ],
  [ Win_Shift_L, Win_6 ],
  [ Win_Shift_L, Win_minus ],
  [ Win_Shift_L, Win_slash ],
  [ Win_Shift_L, Win_1 ],
  [ Win_Shift_L, Win_7 ],
  [ Win_Shift_L, Win_backslash ],
  [ Win_Shift_L, Win_comma ],
  [ Win_Shift_L, Win_period ],
  [ Win_quoteleft ],
  [ Win_backslash ]]

# | | | | | | | | | | | | | | | | | | | "
KeyMap[0][5] = [ 
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ]]

# {F11}|{F2}|{F1}| |{F3}| |{F4}| |{F5}| |{F12}|{F7}| |{F6}| |{F8}| |{F9}| |{F10}"
KeyMap[0][6] = [ 
  [ Win_F11 ],
  [ Win_F2 ],
  [ Win_F1 ],
  [  ],
  [ Win_F3 ],
  [  ],
  [ Win_F4 ],
  [  ],
  [ Win_F5 ],
  [  ],
  [ Win_F12 ],
  [ Win_F7 ],
  [  ],
  [ Win_F6 ],
  [  ],
  [ Win_F8 ],
  [  ],
  [ Win_F9 ],
  [  ],
  [ Win_F10 ]]


# | | | | | | | | | | | | | | | | | | | "
KeyMap[0][7] = [ 
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ]]


# g|r|f|t|d|e|w|s|q|a|m|h|n|u|i|l|o|c|p|b"
KeyMap[1][0] = [ 
  [ Win_Shift_L, Win_G],
  [ Win_Shift_L, Win_R],
  [ Win_Shift_L, Win_F],
  [ Win_Shift_L, Win_T],
  [ Win_Shift_L, Win_D],
  [ Win_Shift_L, Win_E],
  [ Win_Shift_L, Win_W],
  [ Win_Shift_L, Win_S],
  [ Win_Shift_L, Win_Q],
  [ Win_Shift_L, Win_A],
  [ Win_Shift_L, Win_M],
  [ Win_Shift_L, Win_H],
  [ Win_Shift_L, Win_N],
  [ Win_Shift_L, Win_U],
  [ Win_Shift_L, Win_I],
  [ Win_Shift_L, Win_L],
  [ Win_Shift_L, Win_O],
  [ Win_Shift_L, Win_C],
  [ Win_Shift_L, Win_P],
  [ Win_Shift_L, Win_B]]


# {up}|{down}|{left}|{right}|{tab}|{space}|:|{del}|@|{esc}|{pgup}|{pgdn}|{home}|{end}|{enter}|{bs}|.|;|,|{printscreen}"
KeyMap[1][1] = [ 
  [ Win_Shift_L, Win_Up],
  [ Win_Shift_L, Win_Down],
  [ Win_Shift_L, Win_Left],
  [ Win_Shift_L, Win_Right],
  [ Win_Shift_L, Win_Tab],
  [ Win_Shift_L, Win_space],
  [ Win_Shift_L, Win_colon],
  [ Win_Shift_L, Win_Delete],
  [ Win_Shift_L, Win_at],
  [ Win_Shift_L, Win_Escape],
  [ Win_Shift_L, Win_Page_Up],
  [ Win_Shift_L, Win_Page_Down],
  [ Win_Shift_L, Win_Home],
  [ Win_Shift_L, Win_End],
  [ Win_Shift_L, Win_Return],
  [ Win_Shift_L, Win_BackSpace],
  [ Win_Shift_L, ],
  [ Win_Shift_L, Win_semicolon],
  [ Win_Shift_L, ],
  [ Win_Shift_L, Win_Print]]


# {#}|2|,|1|{+}|3|-|4|=|5|$|7|6|.|8|*|9|;|0|/"
KeyMap[1][2] = [ 
  [ Win_numbersign ],
  [ Win_2 ],
  [ Win_comma ],
  [ Win_1 ],
  [ Win_Shift_L, Win_equal ],
  [ Win_3 ],
  [ Win_minus ],
  [ Win_4 ],
  [ Win_equal ],
  [ Win_5 ],
  [ Win_Shift_L, Win_4 ],
  [ Win_7 ],
  [ Win_6 ],
  [ Win_period ],
  [ Win_8 ],
  [ Win_Shift_L, Win_8 ],
  [ Win_9 ],
  [ Win_semicolon ],
  [ Win_0 ],
  [ Win_slash ]]

#  | | |j||v| |y| || | |k||x| |z|| | "
KeyMap[1][3] = [ 
  [  ],
  [  ],
  [  ],
  [ Win_Shift_L, Win_J ],
  [  ],
  [ Win_Shift_L, Win_V ],
  [  ],
  [ Win_Shift_L, Win_Y ],
  [  ],
  [  ],
  [  ],
  [  ],
  [ Win_Shift_L, Win_K ],
  [  ],
  [ Win_Shift_L, Win_X ],
  [  ],
  [ Win_Shift_L, Win_Z ],
  [  ],
  [  ],
  [  ]]

# ~,%,"",',(,),{{},{}},[,],{^},_,?,!,&,|,<,>,``,\"
KeyMap[1][4] = [ 
  [ Win_Shift_L, Win_numbersign ],
  [ Win_Shift_L, Win_5 ],
  [ Win_Shift_L, Win_2 ],
  [ Win_apostrophe],
  [ Win_Shift_L, Win_9 ],
  [ Win_Shift_L, Win_0 ],
  [ Win_Shift_L, Win_bracketleft ],
  [ Win_Shift_L, Win_bracketright ],
  [ Win_bracketleft  ],
  [ Win_bracketright ],
  [ Win_Shift_L, Win_6 ],
  [ Win_Shift_L, Win_minus ],
  [ Win_Shift_L, Win_slash ],
  [ Win_Shift_L, Win_1 ],
  [ Win_Shift_L, Win_7 ],
  [ Win_Shift_L, Win_backslash ],
  [ Win_Shift_L, Win_comma ],
  [ Win_Shift_L, Win_period ],
  [ Win_quoteleft ],
  [ Win_backslash ]]

#  | | | | | | | | | | | | | | | | | | | "
KeyMap[1][5] = [ 
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ]]

# {F11}|{F2}|{F1}| |{F3}| |{F4}| |{F5}| |{F12}|{F7}| |{F6}| |{F8}| |{F9}| |{F10}"
KeyMap[1][6] = [ 
  [ Win_Shift_L, Win_F11 ],
  [ Win_Shift_L, Win_F2 ],
  [ Win_Shift_L, Win_F1 ],
  [  ],
  [ Win_Shift_L, Win_F3 ],
  [  ],
  [ Win_Shift_L, Win_F4 ],
  [  ],
  [ Win_Shift_L, Win_F5 ],
  [  ],
  [ Win_Shift_L, Win_F12 ],
  [ Win_Shift_L, Win_F7 ],
  [  ],
  [ Win_Shift_L, Win_F6 ],
  [  ],
  [ Win_Shift_L, Win_F8 ],
  [  ],
  [ Win_Shift_L, Win_F9 ],
  [  ],
  [ Win_Shift_L, Win_F10 ]]

# | | | | | | | | | | | | | | | | | | | "
KeyMap[1][7] = [ 
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ],
  [  ]]

##################################### state ####################################
# State of the colour shift keys. Index 0 is for the left hand side keys.
$colorshift=[0, 0]

# State of the normal shift key, bit 0 is for left shift.
$shift=0

##################################### action ###################################

# Send the key presses in the list.
def process_keydn_list( list)
  for i in 0...list.length
    UHK_keydn( list[i])
  end
end

# Send the key releases in the list in reverse order.
def process_keyup_list( list)
  n=list.length
  n1=n-1
  for i in 0...n
    UHK_keyup( list[n1-i])
  end
end

# Handle the start and stop marks.
def UHK_USB_mark( code, millisecs)
  # print "Mark: #{code}, #{millisecs}\n"
end

# Handle key press events.
def UHK_USB_keydn( code)
  # print "Down #{code}\n"
 
  if code == USB_PrintScreen
    exit(1)
  end

  # Check if a shift key is pressed.
  if code == USB_SHIFT_L
    $shift=$shift | 1
    return
  end
  if code == USB_SHIFT_R
    $shift=$shift | 2
    return
  end

  # Flatten the shift state into a single bit
  shiftInd=0
  if not ($shift == 0)
    shiftInd=1
  end

  # Simple keys can be mapped directly from USB code to X11 code. There are
  # some keys that are to be ignored.
  if SimpleKeys[shiftInd].has_key?( code)
    # print " simple  -> #{SimpleKeys[$shift][code]}\n"
    process_keydn_list( SimpleKeys[shiftInd][code])
    return
  end
  #  Set the resective bits of the colorshift variables and update the overlay
  #  image.
  if ColorKeys.has_key?(code)
    $colorshift[ColorKeys[code][0]] = $colorshift[ColorKeys[code][0]] | ColorKeys[code][1]
    # print "  color #{$colorshift}\n"
    cs=$colorshift[0] | $colorshift[1]
    return
  end

  # Map the backside keys to their IDs and map those IDs to the X11 codes.
  # First handle the special case of shift-, send as ;. Turn it into shift-,
  if code==USB_Semicolon
    # print "Shift: #{shiftInd}\n"
    if (shiftInd==0) 
      code=USB_Comma
      shiftInd=1
    else
      code=USB_Period
    end
  end

  if BacksideId.has_key?(code)
    id=BacksideId[code]
    cs=$colorshift[0] | $colorshift[1]
    process_keydn_list( KeyMap[shiftInd][cs][id])
  end
end

def UHK_USB_keyup( code)
  # Shift
  # print "Up #{code}\n"
  if code == USB_SHIFT_L
    $shift=$shift & 2
    return
  end
  if code == USB_SHIFT_R
    $shift=$shift & 1
    return
  end

  shiftInd=0
  if not ($shift == 0)
    shiftInd=1
  end

  # Simple keys
  if SimpleKeys[shiftInd].has_key?( code)
    # print " simple  -> #{SimpleKeys[$shift][code]}\n"
    process_keyup_list( SimpleKeys[shiftInd][code]);
    return
  end
  # Colorshift
  if ColorKeys.has_key?(code)
    $colorshift[ColorKeys[code][0]] = $colorshift[ColorKeys[code][0]] & ( ColorKeys[code][1] ^ 7)
    # print "  color #{$colorshift}\n"
    cs=$colorshift[0] | $colorshift[1]
    return
  end
  # Backside keys
  # First handle the special cases
  if code==USB_Semicolon
    if (shiftInd==0) 
      code=USB_Comma
      shiftInd=1
    else
      code=USB_Period
    end
  end
  if BacksideId.has_key?(code)
    id=BacksideId[code]
    cs=$colorshift[0] | $colorshift[1]
    # print " back -> #{ks}\n"
    process_keyup_list( KeyMap[shiftInd][cs][id])
  end
end

# Connect to the alphagrip if present.
UHK_connect( 0x510, 0x100b, 0x81)

print "usbhotkey AG5 script started\n"

